# -*- coding: utf-8 -*-
"""
    bromelia.avps.ietf.rfc6733
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains Diameter AVP classes defined in IETF RFC 6733.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import datetime

from ...__version__ import __version__

from ...base import DiameterAVP
from ...constants.ietf.rfc6733 import *
from ...types import *

PRODUCT_NAME = f"Python bromelia v{__version__}"


class UserNameAVP(DiameterAVP, UTF8StringType):
    """Implementation of User-Name AVP in Section 8.14 of IETF RFC 6733.

    The User-Name AVP (AVP Code 1) [RADIUS] is of type UTF8String.
    """
    code = USER_NAME_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, UserNameAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        UTF8StringType.__init__(self, data=data)


class ClassAVP(DiameterAVP, OctetStringType):
    """Implementation of Class AVP in Section 8.20 of IETF RFC 6733.

    The Class AVP (AVP Code 25) is of type OctetString.
    """
    code = CLASS_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, ClassAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)

        OctetStringType.__init__(self, data=data)


class SessionTimeoutAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Session-Timeout AVP in Section 8.13 of IETF RFC 6733.

    The Session-Timeout AVP (AVP Code 27) is of type Unsigned32.
    """
    code = SESSION_TIMEOUT_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, SessionTimeoutAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class ProxyStateAVP(DiameterAVP, OctetStringType):
    """Implementation of Proxy-State AVP in Section 6.7.4 of IETF RFC 6733.

    The Proxy-State AVP (AVP Code 33) is of type OctetString.
    """
    code = PROXY_STATE_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, ProxyStateAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)    
        OctetStringType.__init__(self, data=data)


class AcctSessionIdAVP(DiameterAVP, OctetStringType):
    """Implementation of Acct-Session-Id AVP in Section 9.8.4 of IETF RFC 6733.

    The Acct-Session-Id AVP (AVP Code 44) is of type OctetString.
    """
    code = ACCT_SESSION_ID_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, AcctSessionIdAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)    
        OctetStringType.__init__(self, data=data)


""" REVIEW """
class AcctMultiSessionIdAVP(DiameterAVP, UTF8StringType):
    """Implementation of Acct-Multi-Session-Id AVP in Section 9.8.5 of 
    IETF RFC 6733.

    The Acct-Multi-Session-Id AVP (AVP Code 50) is of type UTF8String.
    """
    code = ACCT_MULTI_SESSION_ID_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        if isinstance(data, str):
            high = "403292"
            low = "403292"
            optional = "403292"
            data = f"{data};{high};{low};{optional}"
        elif isinstance(data, bytes):
            data = data

        DiameterAVP.__init__(self, AcctMultiSessionIdAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)    
        OctetStringType.__init__(self, data=data)


class EventTimestampAVP(DiameterAVP, TimeType):
    """Implementation of Event-Timestamp AVP in Section 8.21 of IETF RFC 6733.

    The Event-Timestamp AVP (AVP Code 55) is of type Time.
    """
    code = EVENT_TIMESTAMP_AVP_CODE
    vendor_id = None

    def __init__(self, data=datetime.datetime.utcnow()):
        DiameterAVP.__init__(self, EventTimestampAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)    
        TimeType.__init__(self, data=data)


class AcctInterimIntervalAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Acct-Interim-Interval AVP in Section 9.8.2 of 
    IETF RFC 6733.

    The Acct-Interim-Interval AVP (AVP Code 85) is of type Unsigned32.
    """
    code = ACCT_INTERIM_INTERVAL_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, AcctInterimIntervalAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class HostIpAddressAVP(DiameterAVP, AddressType):
    """Implementation of Host-IP-Address AVP in Section 5.3.5 of IETF RFC 6733.

    The Host-IP-Address AVP (AVP Code 257) is of type Address.
    """
    code = HOST_IP_ADDRESS_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, HostIpAddressAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        AddressType.__init__(self, data=data)


class AuthApplicationIdAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Auth-Application-Id AVP in Section 6.8 of 
    IETF RFC 6733.

    The Auth-Application-Id AVP (AVP Code 258) is of type Unsigned32.
    """
    code = AUTH_APPLICATION_ID_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, AuthApplicationIdAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class AcctApplicationIdAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Acct-Application-Id AVP in Section 6.9 of 
    IETF RFC 6733.

    The Acct-Application-Id AVP (AVP Code 259) is of type Unsigned32.
    """
    code = ACCT_APPLICATION_ID_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, AcctApplicationIdAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class VendorIdAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Vendor-Id AVP in Section 5.3.3 of IETF RFC 6733.

    The Vendor-Id AVP (AVP Code 266) is of type Unsigned32.
    """
    code = VENDOR_ID_AVP_CODE
    vendor_id = None

    def __init__(self, data=VENDOR_ID_DEFAULT):
        DiameterAVP.__init__(self, VendorIdAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class VendorSpecificApplicationIdAVP(DiameterAVP, GroupedType):
    """Implementation of Vendor-Specific-Application-Id AVP in Section 6.11 of
    IETF RFC 6733.

    The Vendor-Specific-Application-Id AVP (AVP Code 260) is of type Grouped.
    """
    code = VENDOR_SPECIFIC_APPLICATION_ID_AVP_CODE
    vendor_id = None

    mandatory = {
                    "vendor_id": VendorIdAVP,
    }
    optionals = {
                    "auth_application_id": AuthApplicationIdAVP,
                    "acct_application_id": AcctApplicationIdAVP,
    }

    def __init__(self, data):
        DiameterAVP.__init__(self, VendorSpecificApplicationIdAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        GroupedType.__init__(self, data=data)


class RedirectHostUsageAVP(DiameterAVP, EnumeratedType):
    """Implementation of Redirect-Host-Usage AVP in Section 6.13 of 
    IETF RFC 6733.

    The Redirect-Host-Usage AVP (AVP Code 261) is of type Enumerated.
    """
    code = REDIRECT_HOST_USAGE_AVP_CODE
    vendor_id = None

    values = [
                REDIRECT_HOST_USAGE_DONT_CACHE,
                REDIRECT_HOST_USAGE_ALL_SESSION,
                REDIRECT_HOST_USAGE_ALL_REALM,
                REDIRECT_HOST_USAGE_REALM_AND_APPLICATION,
                REDIRECT_HOST_USAGE_ALL_APPLICATION,
                REDIRECT_HOST_USAGE_ALL_HOST,
                REDIRECT_HOST_USAGE_ALL_USER
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, RedirectHostUsageAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        EnumeratedType.__init__(self, data=data)


class RedirectMaxCacheTimeAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Redirect-Max-Cache-Time AVP in Section 6.14 of 
    IETF RFC 6733.

    The Redirect-Max-Cache-Time AVP (AVP Code 262) is of type Unsigned32.
    """
    code = REDIRECT_MAX_CACHE_TIME_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, RedirectMaxCacheTimeAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class SessionIdAVP(DiameterAVP, UTF8StringType):
    """Implementation of Session-Id AVP in Section 8.8 of IETF RFC 6733.

    The Session-Id AVP (AVP Code 263) is of type UTF8String.
    """
    code = SESSION_ID_AVP_CODE
    vendor_id = None

    diff = datetime.datetime.utcnow() - datetime.datetime(1900, 1, 1, 0, 0, 0)
    init = diff.days*24*60*60 + diff.seconds
    id = 0

    def __init__(self, data):
        if isinstance(data, str):
            high = SessionIdAVP.init
            low = SessionIdAVP.id
            optional = "bromelia"
            data = f"{data};{high};{low};{optional}"
            SessionIdAVP.id += 1
            
        elif isinstance(data, bytes):
            data = data

        DiameterAVP.__init__(self, SessionIdAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        UTF8StringType.__init__(self, data=data)


    @staticmethod
    def _init():
        diff = datetime.datetime.utcnow() - datetime.datetime(1900, 1, 1, 0, 0, 0)
        SessionIdAVP.init = diff.days*24*60*60 + diff.seconds



class OriginHostAVP(DiameterAVP, DiameterIdentityType):
    """Implementation of Origin-Host AVP in Section 6.3 of IETF RFC 6733.

    The Origin-Host AVP (AVP Code 264) is of type DiameterIdentity.
    """
    code = ORIGIN_HOST_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, OriginHostAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterIdentityType.__init__(self, data=data)


class SupportedVendorIdAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Supported-Vendor-Id AVP in Section 5.3.6 of 
    IETF RFC 6733.

    The Supported-Vendor-Id AVP (AVP Code 265) is of type Unsigned32.
    """
    code = SUPPORTED_VENDOR_ID_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, SupportedVendorIdAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class FirmwareRevisionAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Firmware-Revision AVP in Section 5.3.4 of 
    IETF RFC 6733.

    The Firmware-Revision AVP (AVP Code 267) is of type Unsigned32.
    """
    code = FIRMWARE_REVISION_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, FirmwareRevisionAVP.code)
        Unsigned32Type.__init__(self, data=data)


class ResultCodeAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Result-Code AVP in Section 7.1 of IETF RFC 6733.

    The Result-Code AVP (AVP Code 268) is of type Unsigned32.
    """
    code = RESULT_CODE_AVP_CODE
    vendor_id = None

    values = [
                DIAMETER_MULTI_ROUND_AUTH,
                DIAMETER_SUCCESS,
                DIAMETER_LIMITED_SUCCESS,
                DIAMETER_COMMAND_UNSUPPORTED,
                DIAMETER_UNABLE_TO_DELIVER,
                DIAMETER_REALM_NOT_SERVED,
                DIAMETER_TOO_BUSY,
                DIAMETER_LOOP_DETECTED,
                DIAMETER_REDIRECT_INDICATION,
                DIAMETER_APPLICATION_UNSUPPORTED,
                DIAMETER_INVALID_HDR_BITS,
                DIAMETER_INVALID_AVP_BITS,
                DIAMETER_UNKNOWN_PEER,
                DIAMETER_AUTHENTICATION_REJECTED,
                DIAMETER_OUT_OF_SPACE,
                DIAMETER_ELECTION_LOST,
                DIAMETER_AVP_UNSUPPORTED,
                DIAMETER_UNKNOWN_SESSION_ID,
                DIAMETER_AUTHORIZATION_REJECTED,
                DIAMETER_INVALID_AVP_VALUE,
                DIAMETER_MISSING_AVP,
                DIAMETER_RESOURCES_EXCEEDED,
                DIAMETER_CONTRADICTING_AVPS,
                DIAMETER_AVP_NOT_ALLOWED,
                DIAMETER_AVP_OCCURS_TOO_MANY_TIMES,
                DIAMETER_NO_COMMON_APPLICATION,
                DIAMETER_UNSUPPORTED_VERSION,
                DIAMETER_UNABLE_TO_COMPLY,
                DIAMETER_INVALID_BIT_IN_HEADER,
                DIAMETER_INVALID_AVP_LENGTH,
                DIAMETER_INVALID_MESSAGE_LENGTH,
                DIAMETER_INVALID_AVP_BIT_COMBO,
                DIAMETER_NO_COMMON_SECURITY
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, ResultCodeAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class ProductNameAVP(DiameterAVP, UTF8StringType):
    """Implementation of Product-Name AVP in Section 5.3.7 of IETF RFC 6733.

    The Product-Name AVP (AVP Code 269) is of type UTF8String.
    """
    code = PRODUCT_NAME_AVP_CODE
    vendor_id = None

    def __init__(self, data=PRODUCT_NAME):
        DiameterAVP.__init__(self, ProductNameAVP.code)
        UTF8StringType.__init__(self, data=data)


class SessionBindingAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Session-Binding AVP in Section 8.17 of IETF RFC 6733.

    The Session-Binding AVP (AVP Code 270) is of type Unsigned32.
    """
    code = SESSION_BINDING_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, SessionBindingAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class SessionServerFailoverAVP(DiameterAVP, EnumeratedType):
    """Implementation of Session-Server-Failover AVP in Section 8.18 of 
    IETF RFC 6733.

    The Session-Server-Failover AVP (AVP Code 271) is of type Enumerated.
    """
    code = SESSION_SERVER_FAILOVER_AVP_CODE
    vendor_id = None
    
    values = [
                SESSION_SERVER_FAILOVER_REFUSE_SERVICE,
                SESSION_SERVER_FAILOVER_TRY_AGAIN,
                SESSION_SERVER_FAILOVER_ALLOW_SERVICE,
                SESSION_SERVER_FAILOVER_TRY_AGAIN_ALLOW_SERVICE
    ]

    def __init__(self, data=SESSION_SERVER_FAILOVER_REFUSE_SERVICE):
        DiameterAVP.__init__(self, SessionServerFailoverAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        EnumeratedType.__init__(self, data=data)


class MultiRoundTimeOutAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Multi-Round-Time-Out AVP in Section 8.19 of 
    IETF RFC 6733.

    The Multi-Round-Time-Out AVP (AVP Code 272) is of type Unsigned32.
    """
    code = MULTI_ROUND_TIME_OUT_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, MultiRoundTimeOutAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class DisconnectCauseAVP(DiameterAVP, EnumeratedType):
    """Implementation of Disconnect-Cause AVP in Section 5.4.3 of 
    IETF RFC 6733.

    The Disconnect-Cause AVP (AVP Code 273) is of type Enumerated.
    """
    code = DISCONNECT_CAUSE_AVP_CODE
    vendor_id = None
    
    values = [
                DISCONNECT_CAUSE_REBOOTING,
                DISCONNECT_CAUSE_BUSY,
                DISCONNECT_CAUSE_DO_NOT_WANT_TO_TALK_TO_YOU
    ]

    def __init__(self, data=DISCONNECT_CAUSE_REBOOTING):
        DiameterAVP.__init__(self, DisconnectCauseAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        EnumeratedType.__init__(self, data=data)


class AuthRequestTypeAVP(DiameterAVP, EnumeratedType):
    """Implementation of Auth-Request-Type AVP in Section 8.7 of IETF RFC 6733.

    The Auth-Request-Type AVP (AVP Code 274) is of type Enumerated.
    """
    code = AUTH_REQUEST_TYPE_AVP_CODE
    vendor_id = None
    
    values = [
                AUTH_REQUEST_TYPE_AUTHENTICATE_ONLY,
                AUTH_REQUEST_TYPE_AUTHORIZE_ONLY,
                AUTH_REQUEST_TYPE_AUTHORIZE_AUTHENTICATE
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, AuthRequestTypeAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        EnumeratedType.__init__(self, data=data)


class AuthGracePeriodAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Auth-Grace-Period AVP in Section 8.10 of 
    IETF RFC 6733.

    The Auth-Grace-Period AVP (AVP Code 276) is of type Unsigned32.
    """
    code = AUTH_GRACE_PERIOD_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, AuthGracePeriodAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class AuthSessionStateAVP(DiameterAVP, EnumeratedType):
    """Implementation of Auth-Session-State AVP in Section 8.11 of 
    IETF RFC 6733.

    The Auth-Session-State AVP (AVP Code 277) is of type Enumerated.
    """
    code = AUTH_SESSION_STATE_AVP_CODE
    vendor_id = None

    values = [
                STATE_MAINTAINED,
                NO_STATE_MAINTAINED
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, AuthSessionStateAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        EnumeratedType.__init__(self, data=data)


class OriginStateIdAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Origin-State-Id AVP in Section 8.16 of IETF RFC 6733.

    The Origin-State-Id AVP (AVP Code 278) is of type Unsigned32.
    """
    code = ORIGIN_STATE_ID_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, OriginStateIdAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class FailedAvpAVP(DiameterAVP, GroupedType):
    """Implementation of Failed-AVP AVP in Section 7.5 of IETF RFC 6733.

    The Failed-AVP AVP (AVP Code 279) is of type Grouped.
    """
    code = FAILED_AVP_AVP_CODE
    vendor_id = None

    mandatory = {}
    optionals = {}

    def __init__(self, data):
        DiameterAVP.__init__(self, FailedAvpAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        GroupedType.__init__(self, data=data)


class ProxyHostAVP(DiameterAVP, DiameterIdentityType):
    """Implementation of Proxy-Host AVP in Section 6.7.3 of IETF RFC 6733.

    The Proxy-Host AVP (AVP Code 280) is of type DiameterIdentity.
    """
    code = PROXY_HOST_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, ProxyHostAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterIdentityType.__init__(self, data=data)


class ErrorMessageAVP(DiameterAVP, UTF8StringType):
    """Implementation of Error-Message AVP in Section 7.3 of IETF RFC 6733.

    The Error-Message AVP (AVP Code 281) is of type UTF8String.
    """
    code = ERROR_MESSAGE_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, ErrorMessageAVP.code)
        UTF8StringType.__init__(self, data=data)


class RouteRecordAVP(DiameterAVP, DiameterIdentityType):
    """Implementation of Route-Record AVP in Section 6.7.1 of IETF RFC 6733.

    The Route-Record AVP (AVP Code 282) is of type DiameterIdentity.
    """
    code = ROUTE_RECORD_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, RouteRecordAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterIdentityType.__init__(self, data=data)


class DestinationRealmAVP(DiameterAVP, DiameterIdentityType):
    """Implementation of Destination-Realm AVP in Section 6.6 of IETF RFC 6733.

    The Destination-Realm AVP (AVP Code 283) is of type DiameterIdentity.
    """
    code = DESTINATION_REALM_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, DestinationRealmAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterIdentityType.__init__(self, data=data)


class ProxyInfoAVP(DiameterAVP, GroupedType):
    """Implementation of Proxy-Info AVP in Section 6.7.2 of IETF RFC 6733.

    The Proxy-Info AVP (AVP Code 284) is of type Grouped.
    """
    code = PROXY_INFO_AVP_CODE
    vendor_id = None

    mandatory = {
                    "proxy_host": ProxyHostAVP,
                    "proxy_state": ProxyStateAVP,
    }
    optionals = {}

    def __init__(self, data):
        DiameterAVP.__init__(self, ProxyInfoAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        GroupedType.__init__(self, data=data)


class ReAuthRequestTypeAVP(DiameterAVP, EnumeratedType):
    """Implementation of Re-Auth-Request-Type AVP in Section 8.12 of 
    IETF RFC 6733.

    The Re-Auth-Request-Type AVP (AVP Code 285) is of type Enumerated.
    """
    code = RE_AUTH_REQUEST_TYPE_AVP_CODE
    vendor_id = None

    values = [
                RE_AUTH_REQUEST_TYPE_AUTHORIZE_ONLY,
                RE_AUTH_REQUEST_TYPE_AUTHORIZE_AUTHENTICATE
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, ReAuthRequestTypeAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        EnumeratedType.__init__(self, data=data)


class AccountingSubSessionIdAVP(DiameterAVP, Unsigned64Type):
    """Implementation of Accounting-Sub-Session-Id AVP in Section 9.8.6 of 
    IETF RFC 6733.

    The Accounting-Sub-Session-Id AVP (AVP Code 287) is of type Unsigned64.
    """
    code = ACCOUNTING_SUB_SESSION_ID_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, AccountingSubSessionIdAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned64Type.__init__(self, data=data)


class AuthorizationLifetimeAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Authorization-Lifetime AVP in Section 8.9 of 
    IETF RFC 6733.

    The Authorization-Lifetime AVP (AVP Code 291) is of type Unsigned32.
    """
    code = AUTHORIZATION_LIFETIME_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, AuthorizationLifetimeAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class RedirectHostAVP(DiameterAVP, DiameterURIType):
    """Implementation of Redirect-Host AVP in Section 6.12 of IETF RFC 6733.

    The Redirect-Host AVP (AVP Code 292) is of type DiameterURI.
    """
    code = REDIRECT_HOST_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, RedirectHostAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterURIType.__init__(self, data=data)


class DestinationHostAVP(DiameterAVP, DiameterIdentityType):
    """Implementation of Destination-Host AVP in Section 6.5 of IETF RFC 6733.

    The Destination-Host AVP (AVP Code 293) is of type DiameterIdentity.
    """
    code = DESTINATION_HOST_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, DestinationHostAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterIdentityType.__init__(self, data=data)


class ErrorReportingHostAVP(DiameterAVP, DiameterIdentityType):
    """Implementation of Error-Reporting-Host AVP in Section 7.4 of 
    IETF RFC 6733.

    The Error-Reporting-Host AVP (AVP Code 294) is of type DiameterIdentity.
    """
    code = ERROR_REPORTING_HOST_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, ErrorReportingHostAVP.code)
        DiameterIdentityType.__init__(self, data=data)


class TerminationCauseAVP(DiameterAVP, EnumeratedType):
    """Implementation of Termination-Cause AVP in Section 8.47 of 
    IETF RFC 6733.

    The Termination-Cause AVP (AVP Code 295) is of type Enumerated.
    """
    code = TERMINATION_CAUSE_AVP_CODE
    vendor_id = None

    values = [
                DIAMETER_LOGOUT,
                DIAMETER_SERVICE_NOT_PROVIDED,
                DIAMETER_BAD_ANSWER,
                DIAMETER_ADMINISTRATIVE,
                DIAMETER_LINK_BROKEN,
                DIAMETER_AUTH_EXPIRED,
                DIAMETER_USER_MOVED,
                DIAMETER_SESSION_TIMEOUT
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, TerminationCauseAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        EnumeratedType.__init__(self, data=data)


class OriginRealmAVP(DiameterAVP, DiameterIdentityType):
    """Implementation of Origin-Realm AVP in Section 6.4 of IETF RFC 6733.

    The Origin-Realm AVP (AVP Code 296) is of type DiameterIdentity.
    """
    code = ORIGIN_REALM_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, OriginRealmAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        DiameterIdentityType.__init__(self, data=data)


class ExperimentalResultCodeAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Experimental-Result-Code AVP in Section 7.7 of 
    IETF RFC 6733.

    The Experimental-Result-Code AVP (AVP Code 298) is of type Unsigned32.
    """
    code = EXPERIMENTAL_RESULT_CODE_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, ExperimentalResultCodeAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class ExperimentalResultAVP(DiameterAVP, GroupedType):
    """Implementation of Experimental-Result AVP in Section 7.6 of 
    IETF RFC 6733.

    The Experimental-Result AVP (AVP Code 297) is of type Grouped.
    """
    code = EXPERIMENTAL_RESULT_AVP_CODE
    vendor_id = None

    mandatory = {
                    "vendor_id": VendorIdAVP,
                    "experimental_result_code": ExperimentalResultCodeAVP,
    }
    optionals = {}

    def __init__(self, data):
        DiameterAVP.__init__(self, ExperimentalResultAVP.code)
        GroupedType.__init__(self, data=data)


class InbandSecurityIdAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Inband-Security-Id AVP in Section 6.10 of 
    IETF RFC 6733.

    The Inband-Security-Id AVP (AVP Code 299) is of type Unsigned32.
    """
    code = INBAND_SECURITY_ID_AVP_CODE
    vendor_id = None

    def __init__(self, data):    
        DiameterAVP.__init__(self, InbandSecurityIdAVP.code)
        Unsigned32Type.__init__(self, data=data)


class AccountingRecordTypeAVP(DiameterAVP, EnumeratedType):
    """Implementation of Accounting-Record-Type AVP in Section 9.8.1 of 
    IETF RFC 6733.

    The Accounting-Record-Type AVP (AVP Code 480) is of type Enumerated.
    """
    code = ACCOUNTING_RECORD_TYPE_AVP_CODE
    vendor_id = None
    
    values = [
                ACCOUNTING_RECORD_TYPE_EVENT_RECORD,
                ACCOUNTING_RECORD_TYPE_START_RECORD,
                ACCOUNTING_RECORD_TYPE_INTERIM_RECORD,
                ACCOUNTING_RECORD_TYPE_STOP_RECORD
    ]

    def __init__(self, data=ACCOUNTING_RECORD_TYPE_EVENT_RECORD):
        DiameterAVP.__init__(self, AccountingRecordTypeAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        EnumeratedType.__init__(self, data=data)


class AccountingRealtimeRequiredAVP(DiameterAVP, EnumeratedType):
    """Implementation of Accounting-Realtime-Required AVP in Section 9.8.7 of 
    IETF RFC 6733.

    The Accounting-Realtime-Required AVP (AVP Code 483) is of type Enumerated.
    """
    code = ACCOUNTING_REALTIME_REQUIRED_AVP_CODE
    vendor_id = None
    
    values = [
                ACCOUNTING_REALTIME_REQUIRED_DELIVER_AND_GRANT,
                ACCOUNTING_REALTIME_REQUIRED_GRANT_AND_STORE,
                ACCOUNTING_REALTIME_REQUIRED_GRAND_AND_LOSE
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, AccountingRealtimeRequiredAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        EnumeratedType.__init__(self, data=data)


class AccountingRecordNumberAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Accounting-Record-Number AVP in Section 9.8.3 of 
    IETF RFC 6733.

    The Accounting-Record-Number AVP (AVP Code 485) is of type Unsigned32.
    """
    code = ACCOUNTING_RECORD_NUMBER_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, AccountingRecordNumberAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


# -*- coding: utf-8 -*-
"""
    bromelia.constants.etsi_3gpp.ts_129_329
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains constants defined in ETSI TS 129 329.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..._internal_utils import convert_to_4_bytes


#: Diameter AVPs
MSISDN_AVP_CODE = convert_to_4_bytes(701)



# -*- coding: utf-8 -*-
"""
    bromelia.lib.etsi_3gpp_s6b.messages
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains the Diameter protocol messages for 3GPP S6b
    Application Id.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import platform
import socket

from .avps import *

from ...base import DiameterRequest, DiameterAnswer
from ...constants import *


class AAAnswer(DiameterAnswer):
    """Implementation of AA-Answer (MAA) command as per 
    clause 9.2.2.2.2 of ETSI TS 129.273 V14.3.0 (2017-07).

    The AA-Answer is indicated by the Command Code field set to 265
    and Command Flag's 'R' bit cleared.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_s6b import AAA
        >>> aaa_avps = {
        ...     "mip6_feature_vector": 70368744177664,
        ...     "session_timeout": 10800,
        ... }
        >>> aaa = AAA(**aaa_avps)
        >>> aaa
        <Diameter Message: 265 [AAA] PXY, 16777272 [3GPP S6b], 8 AVP(s)>        
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_application_id": AuthApplicationIdAVP,
                    "auth_request_type": AuthRequestTypeAVP,
                    "result_code": ResultCodeAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
    }
    optionals = { 
                    #"drmp": DrmpAVP,
                    "mip6_feature_vector": Mip6FeatureVectorAVP, 
                    "session_timeout": SessionTimeoutAVP,
                    "apn_configuration": ApnConfigurationAVP,
                    #"qos_resources": QosResourcesAVP,
                    #"an_trusted": AnTrustedAVP,
                    "redirect_host": RedirectHostAVP,
                    #"trace_info": TraceInfoAVP,
                    #"oc_supported_features": OcSupportedFeaturesAVP,
                    #"oc_olr": OcOlrAVP,
                    #"load": LoadAVP,
    }

    def __init__(self,
                 session_id=platform.node(),
                 drmp=None,
                 auth_application_id=DIAMETER_APPLICATION_S6b,
                 auth_request_type=AUTH_REQUEST_TYPE_AUTHORIZE_ONLY,
                 result_code=DIAMETER_SUCCESS,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 mip6_feature_vector=None,
                 session_timeout=None,
                 apn_configuration=None,
                 qos_resources=None,
                 an_trusted=None,
                 redirect_host=None,
                 trace_info=None,
                 oc_supported_features=None,
                 oc_olr=None,
                 load=None,
                 **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=AA_MESSAGE, 
                                application_id=DIAMETER_APPLICATION_S6b)

        DiameterAnswer._load(self, locals())


class AARequest(DiameterRequest):
    """Implementation of AA-Request (AAR) command as per 
    clause 9.2.2.2.1 of ETSI TS 129.273 V14.3.0 (2017-07).

    The AA-Request is indicated by the Command Code field set to 265
    and the 'R' bit set in the Command Flags field.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_s6b import AAR
        >>> aar_avps = {
        ...     "destination_realm": "remote",
        ...     "user_name": "frodo",
        ... }
        >>> aar = AAR(**aar_avps)
        >>> aar
        <Diameter Message: 265 [AAR] REQ|PXY, 16777272 [3GPP S6b], 7 AVP(s)>
    """    

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_application_id": AuthApplicationIdAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "auth_request_type": AuthRequestTypeAVP,
    }
    optionals = {
                    #"drmp": DrmpAVP,
                    "user_name": UserNameAVP,
                    "mip6_agent_info": Mip6AgentInfoAVP,
                    "mip6_feature_vector": Mip6FeatureVectorAVP, 
                    "visited_network_identifier": VisitedNetworkIdentifierAVP,
                    #"qos_capacity": QosCapacityAVP,
                    "service_selection": ServiceSelectionAVP,
                    #"oc_supported_features": OcSupportedFeaturesAVP,
                    #"origination_time_stamp": OriginationTimeStampAVP,
                    #"maximum_wait_time": MaximumWaitTimeAVP,
                    "supported_features": SupportedFeaturesAVP,
                    #"emergency_services": EmergencyServicesAVP,
    }

    def __init__(self,
                 session_id=platform.node(), 
                 drmp=None,
                 auth_application_id=DIAMETER_APPLICATION_S6b,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 destination_realm=None,
                 auth_request_type=AUTH_REQUEST_TYPE_AUTHORIZE_ONLY,
                 user_name=None,
                 mip6_agent_info=None,
                 mip6_feature_vector=None,
                 visited_network_identifier=None,
                 qos_capacity=None,
                 service_selection=None,
                 oc_supported_features=None,
                 origination_time_stamp=None,
                 maximum_wait_time=None,
                 supported_features=None,
                 emergency_services=None,
                 **kwargs):

        DiameterRequest.__init__(self,
                                 command_code=AA_MESSAGE, 
                                 application_id=DIAMETER_APPLICATION_S6b)

        DiameterRequest._load(self, locals())


# -*- coding: utf-8 -*-
"""
    bromelia.constants.ietf.rfc5447
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~    
    
    This module contains constants defined in IETF RFC 5447.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..._internal_utils import convert_to_4_bytes


#: Diameter AVPs
MIP_HOME_AGENT_HOST_AVP_CODE = convert_to_4_bytes(348)
MIP6_AGENT_INFO_AVP_CODE = convert_to_4_bytes(486)



# -*- coding: utf-8 -*-
"""
    bromelia.lib.apps.etsi_3gpp_rx
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    A Python package with APIs for handling 3GPP Rx Application Id in the
    bromelia Python Library context.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from .avps import *

from .messages import AAAnswer as AAA
from .messages import AARequest as AAR

from .messages import AbortSessionAnswer as ASA
from .messages import AbortSessionRequest as ASR

from .messages import ReAuthAnswer as RAA
from .messages import ReAuthRequest as RAR

from .messages import SessionTerminationAnswer as STA
from .messages import SessionTerminationRequest as STR


# -*- coding: utf-8 -*-
"""
    bromelia.avps.ietf
    ~~~~~~~~~~~~~~~~~~

    This module contains Diameter AVP classes defined in ETSI 3GPP specs.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""



# -*- coding: utf-8 -*-
"""
    bromelia.avps.ietf.rfc5447
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains Diameter AVP classes defined in IETF RFC 5447.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..ietf.rfc6733 import DestinationHostAVP
from ..ietf.rfc6733 import DestinationRealmAVP
from ...base import DiameterAVP
from ...constants.ietf.rfc5447 import *
from ...types import *


class MipHomeAgentHostAVP(DiameterAVP, GroupedType):
    """Implementation of MIP-Home-Agent-Host AVP in both Section 7.3.43 of
    ETSI TS 129 272 V15.10.0 (2020-01) and Section 4.2.3 of IETF RFC 5447.

    The MIP-Home-Agent-Host AVP (AVP Code 348) is of type Grouped.
    """
    code = MIP_HOME_AGENT_HOST_AVP_CODE
    vendor_id = None

    mandatory = {
                    "destination_host": DestinationHostAVP,
                    "destination_realm": DestinationRealmAVP,
    }
    optionals = {}

    def __init__(self, data):
        DiameterAVP.__init__(self, MipHomeAgentHostAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        GroupedType.__init__(self, data=data)


class Mip6AgentInfoAVP(DiameterAVP, GroupedType):
    """Implementation of MIP6-Agent-Info AVP in both Section 7.3.45 of
    ETSI TS 129 272 V15.10.0 (2020-01) and Section 4.2.1 of IETF RFC 5447.

    The MIP6-Agent-Info AVP (AVP Code 486) is of type Grouped.
    """
    code = MIP6_AGENT_INFO_AVP_CODE
    vendor_id = None

    mandatory = {}
    optionals = {
                    # "mip_home_agent_address": MipHomeAgentAddressAVP,
                    "mip_home_agent_host": MipHomeAgentHostAVP,
                    # "mip_home_link_prefix": MipHomeLinkPrefixAVP,
    }

    def __init__(self, data):
        DiameterAVP.__init__(self, Mip6AgentInfoAVP.code)
        DiameterAVP.set_mandatory_bit(self, True)
        GroupedType.__init__(self, data=data)


# -*- coding: utf-8 -*-
"""
    bromelia.utils
    ~~~~~~~~~~~~~~

    Defines utility functions that are consumed by the library.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ._internal_utils import convert_to_integer_from_bytes
from .constants import *

FLAG_VENDOR_BIT = FLAG_VENDOR_SPECIFIC_AND_NOT_MANDATORY_AND_NOT_PROTECTED


def is_result_code_family_1xxx(result_code):
    if result_code >= 1001 and result_code < 2000:
        return True
    else:
        return False


def is_result_code_family_2xxx(result_code):
    if result_code >= 2001 and result_code < 3000:
        return True
    else:
        return False


def is_result_code_family_3xxx(result_code):
    if result_code >= 3001 and result_code < 4000:
        return True
    else:
        return False


def is_result_code_family_4xxx(result_code):
    if result_code >= 4001 and result_code < 5000:
        return True
    else:
        return False


def is_result_code_family_5xxx(result_code):
    if result_code >= 5001 and result_code < 6000:
        return True
    else:
        return False


def is_vendor_id(flag):
    return (convert_to_integer_from_bytes(flag) & 
            convert_to_integer_from_bytes(FLAG_VENDOR_BIT) != 0)


def is_base_request(msg):
    if is_cer_message(msg):
        return True

    if is_dwr_message(msg):
        return True

    if is_dpr_message(msg):
        return True


def is_base_answer(msg):
    if is_cea_message(msg):
        return True

    if is_dwa_message(msg):
        return True

    if is_dpa_message(msg):
        return True


def is_dwr_message(msg):
    if (msg.header.is_request() 
            and msg.header.command_code == DEVICE_WATCHDOG_MESSAGE):
        return True
    return False


def is_dwa_message(msg):
    if (not msg.header.is_request() 
            and msg.header.command_code == DEVICE_WATCHDOG_MESSAGE):
        return True
    return False    


def is_dpr_message(msg):
    if (msg.header.is_request() 
            and msg.header.command_code == DISCONNECT_PEER_MESSAGE):
        return True
    return False


def is_dpa_message(msg):
    if (not msg.header.is_request() 
            and msg.header.command_code == DISCONNECT_PEER_MESSAGE):
        return True
    return False


def is_cer_message(msg):
    if (msg.header.is_request() 
            and msg.header.command_code == CAPABILITIES_EXCHANGE_MESSAGE):
        return True
    return False


def is_cea_message(msg):
    if (not msg.header.is_request() 
            and msg.header.command_code == CAPABILITIES_EXCHANGE_MESSAGE):
        return True
    return False


def is_answer_message(msg):
    if (not msg.header.is_request() or 
            (not msg.header.is_request() and msg.header.is_proxiable())):
        return True
    return False


def is_request_message(msg):
    if (msg.header.is_request() or 
            (msg.header.is_request() and msg.header.is_proxiable())):
        return True
    return False


def is_client_mode(association):
    if association.connection.mode == DIAMETER_AGENT_CLIENT_MODE:
        return True
    return False


def is_server_mode(association):
    if association.connection.mode == DIAMETER_AGENT_SERVER_MODE:
        return True
    return False


def is_1xxx_informational(answer):
    if answer.has_avp("result_code_avp"):
        code = answer.result_code_avp.data
        return bytes([m & n for m, n in zip(DIAMETER_ERROR_1XXX, code)]) == DIAMETER_ERROR_1XXX


def is_2xxx_success(answer):
    if answer.has_avp("result_code_avp"):
        code = answer.result_code_avp.data
        return bytes([m & n for m, n in zip(DIAMETER_ERROR_2XXX, code)]) == DIAMETER_ERROR_2XXX


def is_3xxx_failure(answer):
    if answer.has_avp("result_code_avp"):
        code = answer.result_code_avp.data
        return bytes([m & n for m, n in zip(DIAMETER_ERROR_3XXX, code)]) == DIAMETER_ERROR_3XXX


def is_4xxx_failure(answer):
    if answer.has_avp("result_code_avp"):
        code = answer.result_code_avp.data
        return bytes([m & n for m, n in zip(DIAMETER_ERROR_4XXX, code)]) == DIAMETER_ERROR_4XXX


def is_5xxx_failure(answer):
    if answer.has_avp("result_code_avp"):
        code = answer.result_code_avp.data
        return bytes([m & n for m, n in zip(DIAMETER_ERROR_5XXX, code)]) == DIAMETER_ERROR_5XXX


def is_result_code_error(answer):
    if answer.has_avp("result_code_avp"):
        if answer.header.is_error():
            return True
        return False


def convert_to_4_length_bit(input):
    try:
        return "{:04b}".format(int(input))
    except ValueError:
        return None


special_chars = { 
    "*": "1010",
    "#": "1011",
    "a": "1100",
    "b": "1101",
    "c": "1110",
}


def encode_special_chars_to_tbcd(input):
    encoded = special_chars.get(input)
    return encoded if encoded is not None else convert_to_4_length_bit(input)


def get_two_bits(input, offset):
    return input[offset:offset+2]


def transform_bits(bits):
    new_bits = str(encode_special_chars_to_tbcd(bits[0]))
    new_bits += str(encode_special_chars_to_tbcd(bits[1]))    
    return str(int(new_bits, 2))


def is_special_char(bits):
    return any(char in bits for char in special_chars.keys())


def encode_to_tbcd(input):
    offset, output = 0, ""
    input = str(input) if isinstance(input, int) else input
    
    while offset < len(input):
        bits = get_two_bits(input, offset)

        if len(bits) == 2:
            bits = bits[::-1]
            bits = transform_bits(bits) if is_special_char(bits) else bits
            output += bits
            offset += 2
        else:
            bits = "f" + str(bits)
            output += bits
            return output


def decode_from_tbcd(input):
    offset, output = 0, ""

    while offset < len(input):
        bits = get_two_bits(input, offset)

        if "f" not in bits:
            output += bits[::-1]
            offset += 2
        else:
            output += bits[1]
            return output


# -*- coding: utf-8 -*-
"""
    bromelia.lib.etsi_3gpp_swm.messages
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains the Diameter protocol messages for 3GPP SWm 
    Application Id.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import platform
import socket

from .avps import *

from ...base import DiameterRequest, DiameterAnswer
from ...constants import *


class AbortSessionAnswer(DiameterAnswer):
    """Implementation of Abort-Session-Answer (ASA) command as per 
    clause 7.1.2.4 of ETSI TS 129.273 V14.3.0 (2017-07).

    The Abort-Session-Answer is indicated by the Command Code field set to 274
    and Command Flag's 'R' bit cleared.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_swm import ASA
        >>> asa = ASA()
        >>> asa
        <Diameter Message: 274 [ASA] PXY, 16777264 [3GPP SWm], 4 AVP(s)>
    """

    mandatory = { 
                    "session_id": SessionIdAVP,
                    "result_code": ResultCodeAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }
    optionals = { 
                    #"drmp": DrmpAvp,
    }

    def __init__(self,
                session_id=platform.node(),
                drmp=None,
                result_code=DIAMETER_SUCCESS,
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(), 
                **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=ABORT_SESSION_MESSAGE, 
                                application_id=DIAMETER_APPLICATION_SWm)

        DiameterAnswer._load(self, locals())


class AbortSessionRequest(DiameterRequest):
    """Implementation of Abort-Session-Request (ASR) command as per 
    clause 7.1.2.4 of ETSI TS 129.273 V14.3.0 (2017-07).

    The Abort-Session-Request is indicated by the Command Code field set to 274
    and the 'R' bit set in the Command Flags field.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_swm import ASR
        >>> asr_avps = {
        ...     "destination_realm": "remote",
        ...     "destination_host": "peer_node",
        ...     "user_name": "frodo"
        ... }
        >>> asr = ASR(**asr)
        >>> asr
        <Diameter Message: 274 [ASR] REQ|PXY, 16777264 [3GPP SWm], 8 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "destination_host": DestinationHostAVP,
                    "auth_application_id": AuthApplicationIdAVP,
    }
    optionals = {
                    #"drmp": DrmpAVP,
                    "user_name": UserNameAVP,
                    "auth_session_state": AuthSessionStateAVP,
    }

    def __init__(self,
                 session_id=platform.node(), 
                 drmp=None,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 destination_realm=None,
                 destination_host=None,
                 auth_application_id=DIAMETER_APPLICATION_SWm,
                 user_name=None,
                 auth_session_state=NO_STATE_MAINTAINED,
                 **kwargs):

        DiameterRequest.__init__(self,
                                 command_code=ABORT_SESSION_MESSAGE, 
                                 application_id=DIAMETER_APPLICATION_SWm)

        DiameterRequest._load(self, locals())


class DiameterEapAnswer(DiameterAnswer):
    """Implementation of Diameter-EAP-Answer (DEA) command as per 
    clause 7.2.2.1.2 of ETSI TS 129.273 V14.3.0 (2017-07).

    The Diameter-EAP-Answer is indicated by the Command Code field set to 268
    and Command Flag's 'R' bit cleared.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_swm import DEA
        >>> dea = DEA()
        >>> dea
        <Diameter Message: 268 [DEA], PXY SWm, 9 AVP(s)>
    """

    mandatory = { 
                    "session_id": SessionIdAVP,
                    "auth_application_id": AuthApplicationIdAVP,
                    "auth_request_type": AuthRequestTypeAVP,
                    "result_code": ResultCodeAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }
    optionals = { 
                    #"drmp": DrmpAvp,
                    "eap_payload": EapPayloadAVP,
                    "user_name": UserNameAVP,
                    "eap_master_session_key": EapMasterSessionKeyAVP,
                    #"apn_oi_replacement": ApnOiReplacementAVP,
                    "apn_configuration": ApnConfigurationAVP,
                    "mip6_feature_vector": Mip6FeatureVectorAVP,
                    "mobile_node_identifier": MobileNodeIdentifierAVP,
                    #"trace_info": TraceInfoAVP,
                    "subscription_id": SubscriptionIdAVP,
                    "session_timeout": SessionTimeoutAVP,
                    "mip6_agent_info": Mip6AgentInfoAVP,
                    "x3gpp_charging_characteristics": X3gppChargingCharacteristicsAVP,
                    "redirect_host": RedirectHostAVP,
                    "supported_features": SupportedFeaturesAVP,
                    #"oc_supported_features": OcSupportedFeaturesAVP,
                    #"oc_olr": OcOlrAVP,
                    #"load": LoadAVP,
                    #"access_network_info": AccessNetworkInfoAVP,
                    #"user_location_info_time": UserLocationInfoTime,
                    #"emergency_info": EmergencyInfoAVP,
    }

    def __init__(self,
                session_id=platform.node(),
                drmp=None,
                auth_application_id=DIAMETER_APPLICATION_SWm,
                auth_request_type=AUTH_REQUEST_TYPE_AUTHENTICATE_ONLY,
                result_code=DIAMETER_SUCCESS,
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(), 
                eap_payload=None, 
                user_name=None, 
                eap_master_session_key=None,
                apn_oi_replacement=None,
                apn_configuration=None,
                mip6_feature_vector=None,
                mobile_node_identifier=None,
                trace_info=None,
                subscription_id=None,
                session_timeout=None,
                mip6_agent_info=None,
                x3gpp_charging_characteristics=None,
                redirect_host=None,
                supported_features=None,
                oc_supported_features=None,
                oc_olr=None,
                load=None,
                access_network_info=None,
                user_location_info_time=None,
                emergency_info=None,
                **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=DIAMETER_EAP_MESSAGE, 
                                application_id=auth_application_id)

        DiameterAnswer._load(self, locals())


class DiameterEapRequest(DiameterRequest):
    """Implementation of Diameter-EAP-Request (DER) command as per 
    clause 7.2.2.1.1 of ETSI TS 129.273 V14.3.0 (2017-07).

    The Diameter-EAP-Request is indicated by the Command Code field set to 268
    and the 'R' bit set in the Command Flags field.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_swm import DER
        >>> der = DER()
        >>> der
        <Diameter Message: 268 [DER], REQ, PXY SWm, 7 AVP(s)>
    """    

    mandatory = {
                    "session_id": SessionIdAVP,
                    "auth_application_id": AuthApplicationIdAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "auth_request_type": AuthRequestTypeAVP,
                    "eap_payload": EapPayloadAVP
    }
    optionals = {
                    #"drmp": DrmpAvp,
                    "destination_host": DestinationHostAVP,
                    "user_name": UserNameAVP,
                    "rat_type": RatTypeAVP,
                    "service_selection": ServiceSelectionAVP,
                    "mip6_feature_vector": Mip6FeatureVectorAVP,
                    # "qos_capability": QosCapabilityAVP, 
                    "visited_network_identifier": VisitedNetworkIdentifierAVP,
                    # "aaa_failure_indication": AaaFailureIndicationAVP,
                    # "supported_features": SupportedFeaturesAVP,
                    "ue_local_ip_address": UeLocalIpAddressAVP,
                    # "oc_supported_features": OcSupportedFeaturesAVP,
                    # "terminal_information": TerminalInformationAVP,
                    # "emergency_services": EmergencyServicesAVP,
    }

    def __init__(self,
                session_id=platform.node(), 
                drmp=None,
                auth_application_id=DIAMETER_APPLICATION_SWm,
                origin_host=platform.node(), 
                origin_realm=socket.getfqdn(), 
                destination_realm=None,
                destination_host=None,
                auth_request_type=AUTH_REQUEST_TYPE_AUTHENTICATE_ONLY,
                eap_payload=None,
                user_name=None,
                rat_type=None,
                service_selection=None,
                mip6_feature_vector=None,
                qos_capability=None, 
                visited_network_identifier=None,
                aaa_failure_indication=None,
                supported_features=None,
                ue_local_ip_address=None,
                oc_supported_features=None,
                terminal_information=None,
                emergency_services=None,
                **kwargs):

        DiameterRequest.__init__(self, 
                                command_code=DIAMETER_EAP_MESSAGE, 
                                application_id=auth_application_id)

        DiameterRequest._load(self, locals())


# -*- coding: utf-8 -*-
"""
    bromelia.constants.ietf.rfc4072
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~    
    
    This module contains constants defined in IETF RFC 4070.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..._internal_utils import convert_to_1_byte
from ..._internal_utils import convert_to_4_bytes


#: Diameter AVPs
EAP_PAYLOAD_AVP_CODE = convert_to_4_bytes(462)
EAP_MASTER_SESSION_KEY_AVP_CODE = convert_to_4_bytes(464)

#: EAP Codes
EAP_CODE_REQUEST = convert_to_1_byte(1)
EAP_CODE_RESPONSE = convert_to_1_byte(2)
EAP_CODE_SUCCESS = convert_to_1_byte(3)
EAP_CODE_FAILURE = convert_to_1_byte(4)

#: EAP Types
EAP_TYPE_IDENTITY = convert_to_1_byte(1)
EAP_TYPE_UMTS_AUTHENTICATION_AND_KEY_AGREEMENT_EAP = convert_to_1_byte(23)

#: EAP-AKA Subtypes
AKA_CHALLENGE = convert_to_1_byte(1)
NOTIFICATION = convert_to_1_byte(12)


# -*- coding: utf-8 -*-
"""
    bromelia.statemachine
    ~~~~~~~~~~~~~~~~~~~~~

    This module contains the finite state machine defined in Section 5.6 of
    IETF RFC 6733.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import logging
import threading
import time

from ._internal_utils import application_id_look_up
from .config import *
from .process import process_answer
from .process import process_capability_exchange
from .process import process_device_watchdog
from .process import process_request
from .utils import is_answer_message
from .utils import is_client_mode
from .utils import is_request_message
from .utils import is_server_mode
from .utils import is_dwa_message as has_recv_dwa
from .utils import is_dwr_message as has_recv_dwr
from .utils import is_dpa_message as has_recv_dpa
from .utils import is_dpr_message as has_recv_dpr
from .utils import is_cea_message as has_recv_cea
from .utils import is_cer_message as has_recv_cer


""" 
    @ `state` column:
    -> It relies on state classes that SHOULD BE implemented.

    -> State classes have the `run` method and another methods with the 
    `event_` prefixed.
    
    -> `run` method in the entrypoint tracking diameter_connection object. 
    Depeding on the object's state, it triggers its `event_` prefixed methods 
    in order to execute a series of actions.


    @ `event` column:
    -> It relies on state classes methods which begins with `event_` preffix. 
    Each method is triggered by the `run` method whenever the 
    diameter_connection object state changes. Inside the methods a series 
    of actions is implemented.


    @ `action` column:
    -> It relies on actions that SHOULD be taken when a given `event_` prefixed 
    method is called by `run` method. The last step of the actions is define 
    the next state by calling a self.next_state with the updated value.


    @ `next state` column:
    -> It represents the self.next_state value for each state class depending 
    on the event/actions triggered (which in turn depends on 
    diameter_connection object state). The State Machine needs to implement a 
    central attribute which tracks all possible states plus a 
    `get_next_state()` method to get the next state object.
"""

closed_logger = logging.getLogger("Closed")
wait_conn_ack_logger = logging.getLogger("WaitConnAck")
wait_initiator_cea_logger = logging.getLogger("WaitInitiatorCEA")
open_logger = logging.getLogger("Open")
wait_returns_logger = logging.getLogger("WaitReturns")
wait_conn_ack_elect_logger = logging.getLogger("WaitConnAckElect")
closing_logger = logging.getLogger("Closing")

statemachine_logger = logging.getLogger("PeerStateMachine")


class State():
    def __init__(self, diameter_association):
        self.association = diameter_association

    def run(self):
        assert(0, "Function not implemented.")

    def next(self):
        assert(0, "Function not implemented.")


class Closed(State):
    def run(self):
        self.next_state = CLOSED
        self.name = self.next_state

        if is_client_mode(self.association):
            self.event_start()
        
        elif is_server_mode(self.association):
            if not self.association._recv_messages.empty():
                self.association.lock.acquire()
                self.msg = self.association._recv_messages.get()
                self.association.lock.release()

                closed_logger.debug(f"Got message from _recv_messages Queue. "\
                                    f"Now there is/are "\
                                    f"{self.association._recv_messages.qsize()} "\
                                    f"Diameter Message(s).")

                if has_recv_cer(self.msg):
                    self.event_responder_conn_cer()


    def event_start(self):
        closed_logger.debug("Event has been triggered.")

        self.next_state = WAIT_CONN_ACK


    def event_responder_conn_cer(self):
        closed_logger.debug("Event has been triggered.")

        cer = self.msg
        process = process_capability_exchange(self.association, cer)

        if process.is_valid:
            cea = self.association.base.cea
            cea.header.hop_by_hop = cer.header.hop_by_hop
            cea.header.end_to_end = cer.header.end_to_end

            self.association.put_message_into_send_queue(cea)
            self.association.send_message_from_queue()

            self.association.state_is_active = True
            self.next_state = OPEN


class WaitConnAck(State):
    def run(self):
        self.next_state = WAIT_CONN_ACK
        self.name = self.next_state

        if self.association.is_connected():
            if self.association.transport.test_connection():
                self.event_initiator_rcv_conn_ack()  
            else:
                self.event_initiator_rcv_conn_nack()

        #: Processing the Diameter messages received.
        if not self.association._recv_messages.empty():
            self.association.lock.acquire()
            self.msg = self.association._recv_messages.get()
            self.association.lock.release()

            wait_initiator_cea_logger.debug(f"Got message from _recv_messages "\
                                            f"Queue. Now there is/are "\
                                            f"{self.association._recv_messages.qsize()} "\
                                            f"Diameter Message(s).")

            if has_recv_cer(self.msg):
                self.event_responder_conn_cer()            


    def event_initiator_rcv_conn_ack(self):
        wait_conn_ack_logger.debug("Event has been triggered.")

        cer = self.association.base.cer
        self.association.put_message_into_send_queue(cer)
        self.association.send_message_from_queue()

        self.next_state = WAIT_I_CEA


    def event_initiator_rcv_conn_nack(self):
        wait_conn_ack_logger.debug("Event has been triggered.")

        self.next_state = CLOSED


    def event_responder_conn_cer(self):
        wait_conn_ack_logger.debug("Event has been triggered.")

        cer = self.msg
        process = process_capability_exchange(self.association, cer)

        if process.is_valid:
            self.association.state_is_active = True
            self.next_state = WAIT_CONN_ACK_ELECT


    def event_timeout(self):
        """ It needs to be coded """
        wait_conn_ack_logger.debug("Event has been triggered.")

        self.next_state = CLOSED


class WaitInitiatorCEA(State):
    def run(self):
        self.next_state = WAIT_I_CEA
        self.name = self.next_state

        #: Processing the Diameter messages received.
        if not self.association._recv_messages.empty():
            self.association.lock.acquire()
            self.msg = self.association._recv_messages.get()
            self.association.lock.release()

            wait_initiator_cea_logger.debug(f"Got message from _recv_messages "\
                                            f"Queue. Now there is/are "
                                            f"{self.association._recv_messages.qsize()} "\
                                            f"Diameter Message(s).")

            if has_recv_cea(self.msg):
                self.event_open_rcv_cea()            

            elif has_recv_cer(self.msg):
                self.event_responder_conn_cer()

            else:
                self.event_initiator_rcv_non_cea()


    def event_open_rcv_cea(self):
        wait_initiator_cea_logger.debug("Event has been triggered.")

        cea = self.msg
        process = process_capability_exchange(self.association, cea)

        if process.is_valid:           
            self.association.state_is_active = True
            self.next_state = OPEN


    def event_responder_conn_cer(self):
        wait_initiator_cea_logger.debug("Event has been triggered.")

        cer = self.msg
        process = process_capability_exchange(self.association, cer)

        if process.is_valid:
            self.association.state_is_active = True
            self.next_state = WAIT_RETURNS


    def event_initiator_peer_disc(self):
        """ It needs to be coded """
        wait_initiator_cea_logger.debug("Event has been triggered.")

        self.next_state = CLOSED

    
    def event_initiator_rcv_non_cea(self):
        wait_initiator_cea_logger.debug("Event has been triggered.")

        self.next_state = CLOSED


    def event_timeout(self):
        """ It needs to be coded """
        wait_initiator_cea_logger.debug("Event has been triggered.")

        self.next_state = CLOSED


class Open(State):
    def run(self):
        self.next_state = OPEN
        self.name = self.next_state

        self.association.tracking_events()

        if self.association.transport._stop_threads:
            self.event_open_peer_disc()      

        if not self.association.state_is_active:
            self.event_stop()

        if not self.association._send_messages.empty():
            open_logger.debug(f"Got message from _send_messages "\
                              f"Queue. Now there is/are "\
                              f"{self.association._send_messages.qsize()} "\
                              f"Diameter Message(s).")

            self.event_send_message()

        #: Processing the Diameter messages received.
        elif not self.association._recv_messages.empty():
            self.association.lock.acquire()
            self.msg = self.association._recv_messages.get()
            self.association.lock.release()

            open_logger.debug(f"Got message from _recv_messages Queue. Now "\
                              f"there is/are "\
                              f"{self.association._recv_messages.qsize()} "\
                              f"Diameter Message(s).")

            if has_recv_dwr(self.msg):
                self.event_open_rcv_dwr()
            
            elif has_recv_dwa(self.msg):
                self.event_open_rcv_dwa()

            elif has_recv_dpr(self.msg):
                self.event_open_rcv_dpr()
           
            elif has_recv_cer(self.msg):
                self.event_open_rcv_cer()

            elif has_recv_cea(self.msg):
                self.event_open_rcv_cea()

            else:
                self.event_open_rcv_message()
            
            self.msg = None


    def event_send_message(self):
        open_logger.debug("Event has been triggered.")

        self.association.send_message_from_queue()

        self.next_state = OPEN


    def event_open_rcv_message(self):
        open_logger.debug("Event has been triggered.")

        if is_answer_message(self.msg):
            process_answer(self.association, self.msg)
            
        elif is_request_message(self.msg):
            process_request(self.association, self.msg)

        if self.msg.has_avp("user_name_avp"):
            open_logger.debug(f"Message from _recv_messages: "\
                              f"{application_id_look_up(self.msg.header.application_id)[0]}, "\
                              f"{self.msg.header.get_command_code()}, "\
                              f"{self.msg.header.is_request()}, "\
                              f"{self.msg.header.hop_by_hop.hex()}, "\
                              f"LEN: {len(self.msg)}, "\
                              f"USERNAME: {self.msg.user_name_avp.data}")
        else:
            open_logger.debug(f"Message from _recv_messages: "\
                              f"{application_id_look_up(self.msg.header.application_id)[0]}, "\
                              f"{self.msg.header.get_command_code()}, "\
                              f"{self.msg.header.is_request()}, "\
                              f"{self.msg.header.hop_by_hop.hex()}, "\
                              f"LEN: {len(self.msg)}")

        open_logger.debug("Putting into postprocess_recv_messages Queue")

        self.association.postprocess_recv_messages_lock.acquire()
        self.association.postprocess_recv_messages.put(self.msg)
        self.association.postprocess_recv_messages_lock.release()

        self.association.postprocess_recv_messages_ready.set()

        open_logger.debug("Just notified postprocess_recv_messages_ready")
    
        self.next_state = OPEN


    def event_open_rcv_dwr(self):
        open_logger.debug("Event has been triggered.")

        dwr = self.msg
        process = process_device_watchdog(self.association, dwr)

        if process.is_valid:
            dwa = self.association.base.dwa
            dwa.header.hop_by_hop = dwr.header.hop_by_hop
            dwa.header.end_to_end = dwr.header.end_to_end

            self.association.put_message_into_send_queue(dwa)
            self.association.send_message_from_queue()

            self.next_state = OPEN

    
    def event_open_rcv_dwa(self):
        open_logger.debug("Event has been triggered.")

        dwa = self.msg
        process = process_device_watchdog(self.association, dwa)

        if process.is_valid:
            self.next_state = OPEN
        else:
            self.next_state = CLOSING

    
    def event_responder_conn_cer(self):
        open_logger.debug("Event has been triggered.")

        self.next_state = OPEN


    def event_stop(self):
        open_logger.debug("Event has been triggered.")

        dpr = self.association.base.dpr
        self.association.put_message_into_send_queue(dpr)
        self.association.send_message_from_queue()

        self.next_state = CLOSING


    def event_open_rcv_dpr(self):
        open_logger.debug("Event has been triggered.")

        dpr = self.msg
        process = process_device_watchdog(self.association, dpr)

        if process.is_valid:
            dpa = self.association.base.dpa
            dpa.header.hop_by_hop = dpr.header.hop_by_hop
            dpa.header.end_to_end = dpr.header.end_to_end

            self.association.put_message_into_send_queue(dpa)
            self.association.send_message_from_queue()

        time.sleep(SLEEP_TIMER)

        self.association._stop_threads = True
        self.association.postprocess_recv_messages_ready.set()

        self.next_state = CLOSED
        

    def event_open_peer_disc(self):
        open_logger.debug("Event has been triggered.")

        self.next_state = CLOSED


    def event_open_rcv_cer(self):
        open_logger.debug("Event has been triggered.")

        cer = self.msg
        process = process_capability_exchange(self.association, cer)

        if process.is_valid:
            cea = self.association.base.cea
            cea.header.hop_by_hop = cer.header.hop_by_hop
            cea.header.end_to_end = cer.header.end_to_end

            self.association.put_message_into_send_queue(cea)
            self.association.send_message_from_queue()

        self.next_state = OPEN


    def event_open_rcv_cea(self):
        open_logger.debug("Event has been triggered.")

        cea = self.msg
        process = process_capability_exchange(self.association, cea)

        if process.is_valid:
            self.next_state = OPEN


class WaitReturns(State):
    def run(self):
        self.next_state = WAIT_RETURNS
        self.name = self.next_state
    

class WaitConnAckElect(State):
    def run(self):
        self.next_state = WAIT_CONN_ACK_ELECT
        self.name = self.next_state


class Closing(State):   
    def run(self):
        self.next_state = CLOSING
        self.name = self.next_state

        if not self.association._recv_messages.empty():
            self.association.lock.acquire()
            self.msg = self.association._recv_messages.get()
            self.association.lock.release()

            if has_recv_dpa(self.msg):
                closing_logger.debug(f"Got DPA message from _recv_messages "\
                                     f"Queue. Now there is/are "\
                                     f"{self.association._recv_messages.qsize()} "\
                                     f"Diameter Message(s).")
                self.event_rcv_dpa()
           
            else:
                closing_logger.debug(f"Got a non-DPA message from "\
                                     f"_recv_messages Queue. Now there is/are "\
                                     f"{self.association._recv_messages.qsize()} "\
                                     f"Diameter Message(s).")


    def event_rcv_dpa(self):
        open_logger.debug("Event has been triggered.")

        self.association.send_message_from_queue()
        self.association._stop_threads = True

        self.next_state = CLOSED


class PeerStateMachine():
    def __init__(self, diameter_association):
        statemachine_logger.debug("PeerStateMachine has been called.")
        self.association = diameter_association
        self._load_states()


    def _load_states(self):
        statemachine_logger.debug("Loading available states.")
        self.__closed = Closed(self.association)
        self.__wait_conn_ack = WaitConnAck(self.association)
        self.__wait_initiator_cea = WaitInitiatorCEA(self.association)
        self.__open = Open(self.association)
        self.__wait_returns = WaitReturns(self.association)
        self.__wait_conn_ack_elect = WaitConnAckElect(self.association)
        self.__closing = Closing(self.association)

        self.states = {
                        CLOSED: self.__closed,
                        WAIT_CONN_ACK: self.__wait_conn_ack,
                        WAIT_I_CEA: self.__wait_initiator_cea,
                        OPEN: self.__open,
                        WAIT_RETURNS: self.__wait_returns,
                        WAIT_CONN_ACK_ELECT: self.__wait_conn_ack_elect,
                        CLOSING: self.__closing,
        }

        self.current_state = self.states[CLOSED]
        self.is_running = False


    def get_next_state(self, next_state):
        if next_state == CLOSED and self.current_state.name == CLOSED:
            return self.states[CLOSED]

        elif next_state == CLOSED and self.current_state.name != CLOSED:
            self.is_running = False
            self.association.close()

        if next_state in self.states:
            if self.current_state.name != next_state:
                statemachine_logger.info(f"Changing from "\
                                         f"{self.current_state.name.upper()} "\
                                         f"to {next_state.upper()} state.")

            return self.states[next_state]
        else:
            raise TypeError("Input not supported for current state.")        


    def start(self):
        statemachine_logger.debug("Starting PeerStateMachine's thread.")

        threading.Thread(name="peer_state_machine_bootstrapper", 
                         target=self.__start).start()


    def __start(self):
        if not self.is_running:
            self.is_running = True

        while (self.is_running and not self.association.error_has_raised):
            time.sleep(STATE_MACHINE_TICKER)

            self.current_state.run()
            _state = self.current_state.next_state
            self.current_state = self.get_next_state(_state)


    def close(self):
        statemachine_logger.debug("Closing PeerStateMachine's thread.")
        self.association.state_is_active = False


    def get_current_state(self):
        if isinstance(self.current_state, Closed):
            return CLOSED
        elif isinstance(self.current_state, WaitConnAck):
            return WAIT_CONN_ACK
        elif isinstance(self.current_state, WaitInitiatorCEA):
            return WAIT_I_CEA
        elif isinstance(self.current_state, WaitConnAckElect):
            return WAIT_CONN_ACK_ELECT
        elif isinstance(self.current_state, WaitReturns):
            return WAIT_RETURNS
        elif isinstance(self.current_state, Open):
            if is_client_mode(self.association):
                return I_OPEN
            elif is_server_mode(self.association):
                return R_OPEN
        elif isinstance(self.current_state, Closing):
            return CLOSING


import os
import pathlib
import setuptools

from bromelia.__version__ import (
        __title__,
        __version__,
        __license__,
        __author__,
        __author_email__,
        __description__,
        __url__
)

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
with open(os.path.join(HERE, "README.md"), encoding="utf-8", mode="r") as fh:
    long_description = fh.read()

# This call to setup() does all the work
setuptools.setup(
    name=__title__,
    version=__version__,
    license=__license__,
    author=__author__,
    author_email=__author_email__,
    description=__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=__url__,
    download_url=f"https://github.com/heimiricmr/bromelia/releases/tag/v{__version__}",
    packages=setuptools.find_packages(),
    include_package_data=True,
    keywords=[
                "DIAMETER", 
                "3GPP", 
                "EPC", 
                "4G", 
                "IMS", 
                "TELECOM", 
                "TELCO", 
                "RFC6733", 
                "RFC3588", 
                "IETF",
                "VoLTE",
                "VoWiFi"
    ],
    install_requires=[
        "pyyaml"
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Telecommunications Industry",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Communications :: Telephony"
    ],
    python_requires=">=3.7"
)


# -*- coding: utf-8 -*-
"""
    bromelia._internal_utils
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Defines utility functions that are consumed internally by the library.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import datetime
import ipaddress
import logging
import os
import struct
import yaml
import warnings
from collections import namedtuple
from warnings import warn

from .definitions import diameter_application_ids
from .definitions import diameter_avps
from .definitions import diameter_command_codes
from .exceptions import InvalidConfigKey
from .exceptions import InvalidConfigValue


def convert_to_1_byte(content):
    return struct.pack(">B", content)


def convert_to_2_bytes(content):
    return struct.pack(">H", content)


def convert_to_3_bytes(content):
    return content.to_bytes(3, byteorder="big")


def convert_to_4_bytes(content):
    return struct.pack(">L", content)


def convert_to_6_bytes(content):
    return content.to_bytes(6, byteorder="big")


def convert_to_8_bytes(content):
    return struct.pack(">q", content)


def convert_to_integer_from_bytes(integer):
    return int.from_bytes(integer, byteorder="big")


def header_representation(header):
    cmd_code = header.command_code
    application_id = header.application_id
    
    cmd_code_str, cmd_code_int = command_code_look_up(cmd_code)

    flag_representation = ""
    if header.is_request():
        flag_representation += " REQ|"
        cmd_code_str = cmd_code_str[:3]
    else:
        flag_representation += " "
        cmd_code_str = cmd_code_str[4:]

    if header.is_proxiable():
        flag_representation += "PXY|"

    if header.is_error():
        flag_representation += "ERR|"

    application_id = header.application_id
    app_id_str, app_id_int = application_id_look_up(application_id)

    return {
                "cmd_code_str": cmd_code_str,
                "cmd_code_int": cmd_code_int,
                "flag_representation": flag_representation[:-1],
                "app_id_str": app_id_str,
                "app_id_int": app_id_int
    }


def application_id_look_up(application_id):
    if not application_id:
        return "", "Unknown"
    
    for application in diameter_application_ids:
        if application["id"] == convert_to_integer_from_bytes(application_id):
            return application["long_name"], application["id"]
    return "", "Unknown"


def command_code_look_up(command_code):
    if not command_code:
        return "", "Unknown"

    for code in diameter_command_codes:
        if code["id"] == convert_to_integer_from_bytes(command_code):
            return code["short_name"], code["id"]
    return "", "Unknown"


def avp_look_up(avp):
    if not avp.get_vendor_id():
        if avp.get_code() == 0:
            return "Unknown"

        for diameter_avp in diameter_avps:
            if diameter_avp["id"] == avp.get_code():
                return diameter_avp["name"]

    return "Unknown"


def _convert_config_to_connection_obj(config):
    LocalNode = namedtuple("LocalNode", [
                                            "host_name", 
                                            "realm", 
                                            "ip_address", 
                                            "port"
                                        ]
    )
    PeerNode = namedtuple("PeerNode", [
                                            "host_name", 
                                            "realm", 
                                            "ip_address", 
                                            "port"
                                        ]
    )
    Connection = namedtuple("Connection", [
                                            "name", 
                                            "mode",
                                            "local_node",
                                            "peer_node",
                                            "application_ids",
                                            "watchdog_timeout"
                                        ]
    )

    config_mask = [
                    "MODE",
                    "APPLICATIONS",
                    "LOCAL_NODE_HOSTNAME",
                    "LOCAL_NODE_REALM",
                    "LOCAL_NODE_IP_ADDRESS",
                    "LOCAL_NODE_PORT",
                    "PEER_NODE_HOSTNAME",
                    "PEER_NODE_REALM",
                    "PEER_NODE_IP_ADDRESS",
                    "PEER_NODE_PORT",
                    "WATCHDOG_TIMEOUT"
    ]

    for key in config.keys():
        if key not in config_mask:
            raise InvalidConfigKey(f"Invalid config key '{key}' found")

    for key, value in config.items():
        if key == "MODE":
            if value not in ["CLIENT", "SERVER"]:
                raise InvalidConfigValue("Invalid config value '{value}' "\
                                f"found for config key '{key}'. It MUST be "\
                                "either 'CLIENT' or 'SERVER'")

            mode = value

        elif key == "APPLICATIONS":
            if value:
                for app in value:
                    app_keys = app.keys()
                    if not [key for key in app_keys if key in ["vendor_id", "app_id"]]:
                        raise InvalidConfigValue("Invalid config value "\
                                        f"found for config key '{key}'. It "\
                                        "MUST be a dictionary with "\
                                        "'vendor_id' and 'app_id' keys")

                    for key in app_keys:
                        if not isinstance(app[key], bytes):
                            raise InvalidConfigValue("Invalid config value "\
                                            f"'{value}' found for config key "\
                                            f"'{key}'. It MUST be a "\
                                            "dictionary with byte value in "\
                                            "each key")


            application_ids = value

        elif key == "LOCAL_NODE_HOSTNAME":
            local_node_host_name = value
        elif key == "LOCAL_NODE_REALM":
            local_node_realm = value
        elif key == "LOCAL_NODE_IP_ADDRESS":
            try:
                ipaddress.IPv4Address(value)
                local_node_ip_address = value

            except ipaddress.AddressValueError:
                raise InvalidConfigValue(f"Invalid config value '{value}' "\
                                f"found for config key '{key}'. It MUST "\
                                 "correspond to a valid IPv4 address format")

        elif key == "LOCAL_NODE_PORT":
            local_node_port = value

        elif key == "PEER_NODE_HOSTNAME":
            peer_node_host_name = value
        elif key == "PEER_NODE_REALM":
            peer_node_realm = value
        elif key == "PEER_NODE_IP_ADDRESS":
            try:
                ipaddress.IPv4Address(value)
                peer_node_ip_address = value

            except ipaddress.AddressValueError:
                raise InvalidConfigValue(f"Invalid config value '{value}' "\
                                f"found for config key '{key}'. It MUST "\
                                 "correspond to a valid IPv4 address format")

        elif key == "PEER_NODE_PORT":
            peer_node_port = value

        elif key == "WATCHDOG_TIMEOUT":
            if not isinstance(value, int):
                raise InvalidConfigValue(f"Invalid config value '{value}' "\
                                f"found for config key '{key}'. It MUST be "\
                                "'int'")

            watchdog_timeout = value

    local_node = LocalNode(host_name=local_node_host_name,
                           realm=local_node_realm,
                           ip_address=local_node_ip_address,
                           port=local_node_port
    )

    peer_node = PeerNode(host_name=peer_node_host_name,
                         realm=peer_node_realm,
                         ip_address=peer_node_ip_address,
                         port=peer_node_port
    )

    connection = Connection(name="bromelia",
                            mode=mode,
                            application_ids=application_ids,
                            local_node=local_node,
                            peer_node=peer_node,
                            watchdog_timeout=watchdog_timeout)

    return connection


def _convert_file_to_config(filepath=None, variables_dictionary=globals()):
    if not filepath:
        filepath = os.path.join(os.getcwd(), "config.yaml")

    try:
        if os.path.exists(filepath):
            with open(filepath, "r") as config_file:
                from_config_file = yaml.load(config_file, Loader=yaml.FullLoader)

    except Exception as e:
        logging.exception(f"_convert_file_to_config - exception: {e}")

    if from_config_file["api_version"] != "v1":
        raise

    configs = list()

    for spec in from_config_file["spec"]:
        for application in spec["applications"]:
            vendor_id = application["vendor_id"]
            app_id = application["app_id"]

            application["vendor_id"] = variables_dictionary[vendor_id]
            application["app_id"] = variables_dictionary[app_id]

        configs.append({
                            "MODE": spec["mode"].upper(),
                            "APPLICATIONS": spec["applications"],
                            "LOCAL_NODE_HOSTNAME": spec["local"]["hostname"],
                            "LOCAL_NODE_REALM": spec["local"]["realm"],
                            "LOCAL_NODE_IP_ADDRESS": spec["local"]["ip_address"],
                            "LOCAL_NODE_PORT": spec["local"]["port"],
                            "PEER_NODE_HOSTNAME": spec["peer"]["hostname"],
                            "PEER_NODE_REALM": spec["peer"]["realm"],
                            "PEER_NODE_IP_ADDRESS": spec["peer"]["ip_address"],
                            "PEER_NODE_PORT": spec["peer"]["port"],
                            "WATCHDOG_TIMEOUT": 60
        })

    return configs


def get_app_ids(apps):
    text = ""
    for index, app in enumerate(apps):
        app_name = application_id_look_up(app['app_id'])[0]
        text += f"{app_name};"
        if index == (len(apps) - 1):
            return text[:-1]


def get_app_name(filepath=None):
    if not filepath:
        filepath = os.path.join(os.getcwd(), "config.yaml")

    try:
        if os.path.exists(filepath):
            with open(filepath, "r") as config_file:
                from_config_file = yaml.load(config_file, Loader=yaml.FullLoader)

    except Exception as e:
        logging.exception(f"_convert_file_to_config - exception: {e}")

    if from_config_file["api_version"] != "v1":
        raise

    return from_config_file["name"]


def get_logging_filename(app_name=None):
    if app_name is None:
        name = "dsa"
    else:
        if not isinstance(app_name, str):
            name = "dsa"
        else:
            name = app_name.lower()

    now = datetime.datetime.now()

    year = str(now.year).zfill(2)
    month = str(now.month).zfill(2)
    day = str(now.day).zfill(2)

    hour = str(now.hour).zfill(2)
    minute = str(now.minute).zfill(2)
    second = str(now.second).zfill(2)

    return f"log-{name}-"\
           f"{year}-{month}-{day}-{hour}-"\
           f"{minute}-{second}-UTC3-pid_{os.getpid()}.log"


warnings.simplefilter("default")


def show_warn(module, _path, _except=None):
    if module not in ["avps", "messages"]:
        raise ValueError("Deprecation Warning for avps & messages only")

    if _except is None:
        warn(f"Please prefer import Diameter Message classes from "\
             f"bromelia.lib.{_path} instead bromelia.{_path}.{module}. "\
             f"The latter one will be deprecated in the next release",
             DeprecationWarning, stacklevel=5)
    else:
        warn(f"Please prefer import Diameter Message classes from "\
             f"bromelia.lib.{_except} instead bromelia.{_path}.{module}. "\
             f"The latter one will be deprecated in the next release",
             DeprecationWarning, stacklevel=5)


# -*- coding: utf-8 -*-
"""
    bromelia.config
    ~~~~~~~~~~~~~~~

    This module contains configuration structures.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import logging
import os

from ._internal_utils import get_logging_filename

CWD = os.getcwd()

#: Configs for statemachine.py module
STATE_MACHINE_TICKER = 0.0001

CLOSED = "Closed"
WAIT_CONN_ACK = "Wait-Conn-Ack"
WAIT_I_CEA = "Wait-I-CEA"
WAIT_CONN_ACK_ELECT = "Wait-Conn-Ack/Elect"
WAIT_RETURNS = "Wait-Returns"
I_OPEN = "I-Open"
R_OPEN = "R-Open"
OPEN = "Open"
CLOSING = "Closing"

#: Configs for setup.py module
SEND_BUFFER_MAXIMUM_SIZE = 4096*64
LISTENING_TICKER = 0.01
WAITING_CONN_TIMER = 2
SLEEP_TIMER = 4

#: Configs for bromelia.py module
BROMELIA_TICKER = STATE_MACHINE_TICKER
BROMELIA_LOADING_TICKER = 0.1
SEND_THRESHOLD_TICKER = 0.05
PROCESS_TIMER = 0.001
SEND_EVENT_THRESHOLD_TICKER = 0.1

REQUEST_THRESHOLD = 40 #25 #10
ANSWER_THRESHOLD = 40 #25 #10
SEND_THRESHOLD = 50 #40 #30

#: Configs for transport.py module
TRACKING_SOCKET_EVENTS_TIMEOUT = 1


class Config(dict):
    def __init__(self, defaults=None):
        dict.__init__(self, defaults or {})


class DiameterLogging(object):
    def __init__(self, debug=False, is_logging=False, app_name=None):
        if debug:
            LOGGING_LEVEL = logging.DEBUG
        else:
            LOGGING_LEVEL = logging.INFO

        LOGGING_FORMAT = "%(asctime)s [%(levelname)s] [%(process)d] "\
                        "[%(thread)d:%(threadName)s] %(module)s [%(name)s] "\
                        "[%(funcName)s]: %(message)s"

        filename = get_logging_filename(app_name)
        LOGGING_PATH = os.path.join(CWD, filename)

        if is_logging:
            logging.basicConfig(level=LOGGING_LEVEL,
                                format=LOGGING_FORMAT,
                                filename=LOGGING_PATH,
                                filemode="a")



# -*- coding: utf-8 -*-
"""
    bromelia.etsi_3gpp.ts_129_061
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains Diameter AVP classes defined in ETSI TS 129 061.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ...base import DiameterAVP
from ...constants.etsi_3gpp.ts_129_061 import *
from ...types import *


class X3gppChargingCharacteristicsAVP(DiameterAVP, UTF8StringType):
    """Implementation of 3GPP-Charging-Characteristics AVP in Section 16.4.7.2 
    of ETSI TS 129 061 V10.11.0 (2014-10).

    The 3GPP-Charging-Characteristics AVP (AVP Code 13) is of type UTF8String.
    """
    code = X_3GPP_CHARGING_CHARACTERISTICS_AVP_CODE
    vendor_id = VENDOR_ID_3GPP

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             X3gppChargingCharacteristicsAVP.code,
                             X3gppChargingCharacteristicsAVP.vendor_id)
        DiameterAVP.set_vendor_id_bit(self, True)
        UTF8StringType.__init__(self, data=data, vendor_id=VENDOR_ID_3GPP)


# -*- coding: utf-8 -*-
"""
    bromelia.lib.etsi_3gpp_swx.messages
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains the Diameter protocol messages for 3GPP SWx
    Application Id.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

import platform
import socket

from .avps import *

from ...base import DiameterRequest, DiameterAnswer
from ...constants import *


class MultimediaAuthAnswer(DiameterAnswer):
    """Implementation of Multimedia-Auth-Answer (MAA) command as per 
    clause 8.2.2.2 of ETSI TS 129.273 V14.3.0 (2017-07).

    The Multimedia-Auth-Answer is indicated by the Command Code field set to 
    303 and Command Flag's 'R' bit cleared.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_swx import MAA
        >>> maa = MAA()
        >>> maa
        <Diameter Message: 303 [MAA], PXY SWx, 9 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "user_name": UserNameAVP,
    }
    optionals = { 
                    #"drmp": DrmpAVP,
                    "result_code": ResultCodeAVP,
                    "experimental_result": ExperimentalResultAVP,
                    "sip_number_auth_items": SipNumberAuthItemsAVP,
                    "sip_auth_data_item": SipAuthDataItemAVP,
                    #"x3gpp_aaa_server_name": X3gppAaaServerNameAVP,
                    #"oc_supported_features": OcSupportedFeaturesAVP,
                    #"oc_olr": OcOlrAVP,
                    #"load": LoadAVP,
                    "supported_features": SupportedFeaturesAVP,
    }

    def __init__(self,
                 session_id=platform.node(),
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_SWx)],
                 result_code=None,
                 experimental_result=None,
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 user_name=None,
                 sip_number_auth_items=None,
                 sip_auth_data_item=None,
                 x3gpp_aaa_server_name=None,
                 oc_supported_features=None,
                 oc_olr=None,
                 load=None,
                 supported_features=None,
                 **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=MULTIMEDIA_AUTH_MESSAGE, 
                                application_id=DIAMETER_APPLICATION_SWx)

        DiameterAnswer._load(self, locals())


class MultimediaAuthRequest(DiameterRequest):
    """Implementation of Multimedia-Auth-Request (MAR) command as per 
    clause 8.2.2.1 of ETSI TS 129.273 V14.3.0 (2017-07).

    The Multimedia-Auth-Request is indicated by the Command Code field set to 
    303 and the 'R' bit set in the Command Flags field.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_swx import MAR
        >>> mar = MAR()
        >>> mar
        <Diameter Message: 303 [MAR], REQ, PXY SWx, 10 AVP(s)>
    """    

    mandatory = {
                    "session_id": SessionIdAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "user_name": UserNameAVP,
                    "sip_auth_data_item": SipAuthDataItemAVP,
                    "sip_number_auth_items": SipNumberAuthItemsAVP
    }
    optionals = {
                    #"drmp": DrmpAVP,
                    "destination_host": DestinationHostAVP,
                    "rat_type": RatTypeAVP,
                    #"anid": AnidAVP,
                    "visited_network_identifier": VisitedNetworkIdentifierAVP,
                    "terminal_information": TerminalInformationAVP,
                    #"aaa_failure_indication": AaaFailureIndicationAVP,
                    #"oc_supported_features": OcSupportedFeaturesAVP,
                    "supported_features": SupportedFeaturesAVP
    }

    def __init__(self,
                 session_id=platform.node(), 
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_SWx)],
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 destination_realm=None,
                 destination_host=None,
                 user_name=None,
                 rat_type=None,
                 anid=None,
                 visited_network_identifier=None,
                 terminal_information=None,
                 sip_auth_data_item=[SipAuthenticationSchemeAVP("EAP-AKA")],
                 sip_number_auth_items=5,
                 aaa_failure_indication=None,
                 oc_supported_features=None,
                 supported_features=None,
                 **kwargs):

        DiameterRequest.__init__(self,
                                command_code=MULTIMEDIA_AUTH_MESSAGE, 
                                application_id=DIAMETER_APPLICATION_SWx)

        DiameterRequest._load(self, locals())


class RegistrationTerminationAnswer(DiameterAnswer):
    """Implementation of Registration-Termination-Answer (RTA) command as per 
    clause 8.2.2.4 of ETSI TS 129.273 V14.3.0 (2017-07).

    The Server-Assignment-Answer is indicated by the Command Code field set to 
    304 and Command Flag's 'R' bit cleared.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_swx import RTA
        >>> rta_avps = {
        ...     "user_name": "frodo"
        ... }
        >>> rta = RTA(**rta_avps)
        >>> rta
        <Diameter Message: 304 [RTA], PXY, 16777265 [3GPP SWx], 6 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
    }
    optionals = { 
                    #"drmp": DrmpAVP,
                    "result_code": ResultCodeAVP,
                    "experimental_result": ExperimentalResultAVP,
                    "supported_features": SupportedFeaturesAVP,
    }

    def __init__(self,
                 session_id=platform.node(),
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_SWx)],
                 result_code=None,
                 experimental_result=None,
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 supported_features=None,
                 **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=REGISTRATION_TERMINATION_MESSAGE, 
                                application_id=DIAMETER_APPLICATION_SWx)

        DiameterAnswer._load(self, locals())


class RegistrationTerminationRequest(DiameterRequest):
    """Implementation of Registration-Termination-Request (RTR) command as per 
    clause 8.2.2.4 of ETSI TS 129.273 V14.3.0 (2017-07).

    The Registration-Termination-Request is indicated by the Command Code 
    field set to 304 and the 'R' bit set in the Command Flags field.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_swx import RTR
        >>> rtr_avps = {
        ...     "destination_realm": "example.com",
        ...     "user_name": "frodo"
        ...     "destination_host": "host.example.com",
        ... }
        >>> rtr = RTR(**rtr_avps)
        >>> rtr
        <Diameter Message: 304 [RTR] REQ|PXY, 16777265 [3GPP SWx], 9 AVP(s)>
    """    

    mandatory = {
                    "session_id": SessionIdAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_host": DestinationHostAVP,
                    "destination_realm": DestinationRealmAVP,
                    "user_name": UserNameAVP,
                    "deregistration_reason": DeregistrationReasonAVP
    }
    optionals = {
                    #"drmp": DrmpAVP,
                    "supported_features": SupportedFeaturesAVP,
    }

    def __init__(self, 
                 session_id=platform.node(), 
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_SWx)],
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 destination_host=None,
                 destination_realm=None,
                 user_name=None,
                 deregistration_reason=[ReasonCodeAVP(REASON_CODE_PERMANENT_TERMINATION)],
                 supported_features=None,
                 **kwargs):

        DiameterRequest.__init__(self, 
                                 command_code=REGISTRATION_TERMINATION_MESSAGE, 
                                 application_id=DIAMETER_APPLICATION_SWx)

        DiameterRequest._load(self, locals())


class ServerAssignmentAnswer(DiameterAnswer):
    """Implementation of Server-Assignment-Answer (SAA) command as per 
    clause 8.2.2.3 of ETSI TS 129.273 V14.3.0 (2017-07).

    The Server-Assignment-Answer is indicated by the Command Code field set to 
    301 and Command Flag's 'R' bit cleared.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_swx import SAA
        >>> saa_avps = {
        ...     "user_name": "frodo"
        ... }
        >>> saa = SAA(**sar_avps)
        >>> saa
        <Diameter Message: 301 [SAA], PXY SWx, 8 AVP(s)>
    """

    mandatory = {
                    "session_id": SessionIdAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "user_name": UserNameAVP,
    }
    optionals = { 
                    #"drmp": DrmpAVP,
                    "result_code": ResultCodeAVP,
                    "experimental_result": ExperimentalResultAVP,
                    "non3gpp_user_data": Non3gppUserDataAVP,
                    #"x3gpp_aaa_server_name": X3gppAaaServerNameAVP,
                    #"oc_supported_features": OcSupportedFeaturesAVP,
                    #"oc_olr": OcOlrAVP,
                    #"load": LoadAVP,
                    "supported_features": SupportedFeaturesAVP,
    }

    def __init__(self,
                 session_id=platform.node(),
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_SWx)],
                 result_code=None,
                 experimental_result=None,
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 user_name=None,
                 non3gpp_user_data=None,
                 x3gpp_aaa_server_name=None,
                 oc_supported_features=None,
                 oc_olr=None,
                 load=None,
                 supported_features=None,
                 **kwargs):

        DiameterAnswer.__init__(self, 
                                command_code=SERVER_ASSIGNMENT_MESSAGE, 
                                application_id=DIAMETER_APPLICATION_SWx)

        DiameterAnswer._load(self, locals())


class ServerAssignmentRequest(DiameterRequest):
    """Implementation of Server-Assignment-Request (MAR) command as per 
    clause 8.2.2.3 of ETSI TS 129.273 V14.3.0 (2017-07).

    The Server-Assignment-Request is indicated by the Command Code field set to
    301 and the 'R' bit set in the Command Flags field.

    Usage::

        >>> from bromelia.lib.etsi_3gpp_swx import SAR
        >>> sar_avps = {
        ...     "destination_realm": "example.com",
        ...     "user_name": "frodo"
        ...     "destination_host": "host.example.com",
        ... }
        >>> sar = SAR(**sar_avps)
        >>> sar
        <Diameter Message: 301 [SAR], REQ, PXY SWx, 9 AVP(s)>
    """    

    mandatory = {
                    "session_id": SessionIdAVP,
                    "vendor_specific_application_id": VendorSpecificApplicationIdAVP,
                    "auth_session_state": AuthSessionStateAVP,
                    "origin_host": OriginHostAVP,
                    "origin_realm": OriginRealmAVP,
                    "destination_realm": DestinationRealmAVP,
                    "user_name": UserNameAVP,
                    "server_assignment_type": ServerAssignmentTypeAVP
    }
    optionals = {
                    #"drmp": DrmpAVP,
                    "destination_host": DestinationHostAVP,
                    "service_selection": ServiceSelectionAVP,
                    "context_identifier": ContextIdentifierAVP,
                    "mip6_agent_info": Mip6AgentInfoAVP,
                    "visited_network_identifier": VisitedNetworkIdentifierAVP,
                    #"active_apn": ActiveApnAVP,
                    #"oc_supported_features": OcSupportedFeaturesAVP,
                    "supported_features": SupportedFeaturesAVP,
                    "terminal_information": TerminalInformationAVP,
                    #"emergency_services": EmergencyServicesAVP
    }

    def __init__(self, 
                 session_id=platform.node(), 
                 drmp=None,
                 vendor_specific_application_id=[VendorIdAVP(VENDOR_ID_3GPP), AuthApplicationIdAVP(DIAMETER_APPLICATION_SWx)],
                 auth_session_state=NO_STATE_MAINTAINED,
                 origin_host=platform.node(), 
                 origin_realm=socket.getfqdn(), 
                 destination_host=None,
                 destination_realm=None,
                 service_selection=None,
                 context_identifier=None,
                 mip6_agent_info=None,
                 visited_network_identifier=None,
                 user_name=None,
                 server_assignment_type=SERVER_ASSIGNMENT_TYPE_NO_ASSIGNMENT,
                 active_apn=None,
                 oc_supported_features=None,
                 terminal_information=None,
                 emergency_services=None,
                 **kwargs):

        DiameterRequest.__init__(self, 
                                 command_code=SERVER_ASSIGNMENT_MESSAGE, 
                                 application_id=DIAMETER_APPLICATION_SWx)

        DiameterRequest._load(self, locals())


# -*- coding: utf-8 -*-
"""
    bromelia.constants.etsi_3gpp.ts_129_273
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains constants defined in ETSI TS 129 273.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ..._internal_utils import convert_to_3_bytes
from ..._internal_utils import convert_to_4_bytes


#: Diameter Messages
RE_AUTH_MESSAGE = convert_to_3_bytes(258)
AA_MESSAGE = convert_to_3_bytes(265)
DIAMETER_EAP_MESSAGE = convert_to_3_bytes(268)
SESSION_TERMINATION_MESSAGE = convert_to_3_bytes(275)
SERVER_ASSIGNMENT_MESSAGE = convert_to_3_bytes(301)
MULTIMEDIA_AUTH_MESSAGE = convert_to_3_bytes(303)
REGISTRATION_TERMINATION_MESSAGE = convert_to_3_bytes(304)
PUSH_PROFILE_MESSAGE = convert_to_3_bytes(305)

#: Diameter AVPs
MIP6_FEATURE_VECTOR_AVP_CODE = convert_to_4_bytes(124)
MOBILE_NODE_IDENTIFIER_AVP_CODE = convert_to_4_bytes(506)
NON_3GPP_USER_DATA_AVP_CODE = convert_to_4_bytes(1500)
NON_3GPP_IP_ACCESS_AVP_CODE = convert_to_4_bytes(1501)
NON_3GPP_IP_ACCESS_APN_AVP_CODE = convert_to_4_bytes(1502)
AN_TRUSTED_AVP_CODE = convert_to_4_bytes(1503)

#: List of AN-Trusted AVP values.
#: For more information, please refer to Section 5.2.3.9 of 
#: ETSI TS 129 273 V12.5.0 (2014-10).
AN_TRUSTED_TRUSTED = convert_to_4_bytes(0)
AN_TRUSTED_UNTRUSTED = convert_to_4_bytes(1)

#: List of Non-3GPP-IP-Access values.
#: For more information, please refer to Section 8.2.3.3 of
#  ETSI TS 129 273 V14.5.0 (2019-10).
NON_3GPP_SUBSCRIPTION_ALLOWED = convert_to_4_bytes(0)
NON_3GPP_SUBSCRIPTION_BARRED = convert_to_4_bytes(1)

#: List of Non-3GPP-IP-Access-APN AVP values.
#: For more information, please refer to Section 8.2.3.4 of 
#: ETSI TS 129 273 V14.3.0 (2017-07).
NON_3GPP_APNS_ENABLE = convert_to_4_bytes(0)
NON_3GPP_APNS_DISABLE = convert_to_4_bytes(1)

#: List of Server-Assignment-Type AVP values.
#: For more information, please refer to Section 8.2.3.12 of 
#: ETSI TS 129 273 V14.3.0 (2017-07).
SERVER_ASSIGNMENT_TYPE_NO_ASSIGNMENT = convert_to_4_bytes(0)
SERVER_ASSIGNMENT_TYPE_REGISTRATION = convert_to_4_bytes(1)
SERVER_ASSIGNMENT_TYPE_RE_REGISTRATION = convert_to_4_bytes(2)
SERVER_ASSIGNMENT_TYPE_UNREGISTERED_USER = convert_to_4_bytes(3)
SERVER_ASSIGNMENT_TYPE_TIMEOUT_DEREGISTRATION = convert_to_4_bytes(4)
SERVER_ASSIGNMENT_TYPE_USER_DEREGISTRATION = convert_to_4_bytes(5)
SERVER_ASSIGNMENT_TYPE_DEREGISTRATION_STORE_SERVER_NAME = convert_to_4_bytes(6)
SERVER_ASSIGNMENT_TYPE_USER_DEREGISTRATION_STORE_SERVER_NAME = convert_to_4_bytes(7)
SERVER_ASSIGNMENT_TYPE_ADMINISTRATIVE_DEREGISTRATION = convert_to_4_bytes(8)
SERVER_ASSIGNMENT_TYPE_AUTHENTICATION_FAILURE = convert_to_4_bytes(9)
SERVER_ASSIGNMENT_TYPE_AUTHENTICATION_TIMEOUT = convert_to_4_bytes(10)
SERVER_ASSIGNMENT_TYPE_DEREGISTRATION_TOO_MUCH_DATA = convert_to_4_bytes(11)
SERVER_ASSIGNMENT_TYPE_AAA_USER_DATA_REQUEST = convert_to_4_bytes(12)
SERVER_ASSIGNMENT_TYPE_PGW_UPDATE = convert_to_4_bytes(13)
SERVER_ASSIGNMENT_TYPE_RESTORATION = convert_to_4_bytes(14)



# -*- coding: utf-8 -*-
"""
    bromelia.exceptions
    ~~~~~~~~~~~~~~~~~~~
    
    This module implements the central Diameter application object. It works
    as per Peer State Machine defined in RFC 6733 in order to establish a 
    Diameter association with another Peer Node.

    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""


class DiameterApplicationError(BaseException):
    """A Diameter Application error occurred."""
    pass


class DiameterAssociationError(BaseException):
    """A Diameter Association error occurred."""
    pass


class DiameterMessageError(BaseException):
    """A Diameter Message error occurred."""
    pass


class DiameterHeaderError(BaseException):
    """A Diameter Header error occurred."""
    pass


class DiameterAvpError(BaseException):
    """ A Diameter AVP error occurred."""
    pass


class DiameterTypeError(BaseException):
    """ A Diameter Type error occurred """
    pass


class DiameterHeaderAttributeValueError(BaseException):
    """A valid attribute value is required."""
    pass


class AVPAttributeValueError(BaseException):
    """A valid attribute value is required."""
    pass


class AVPOperationError(BaseException):
    """Invalid operation between two DiameterAVP objects"""


class AVPParsingError(BaseException):
    """An invalid AVP byte stream has been found."""
    pass


class ProcessRequestException(BaseException):
    pass


class DataTypeError(BaseException):
    pass


class ParsingDataTypeError(BaseException):
    pass


class InvalidConfigKey(BaseException):
    """Invalid config key found"""


class InvalidConfigValue(BaseException):
    """Invalid config value found"""


class MissingAttributes(KeyError):
    """ Not found attributes for a given class """


class BromeliaException(BaseException):
    """ Something went wrong in Bromelia class """


class DiameterInvalidAvpValue(BaseException):
    """ Refer to DIAMETER_INVALID_AVP_VALUE constant """

    
class DiameterMissingAvp(BaseException):
    """ Refer to DIAMETER_MISSING_AVP constant"""



# -*- coding: utf-8 -*-
"""
    bromelia.avps.ietf.rfc8506
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains Diameter AVP classes defined in IETF RFC 8506.
    
    :copyright: (c) 2020-present Henrique Marques Ribeiro.
    :license: MIT, see LICENSE for more details.
"""

from ...base import DiameterAVP
from ...constants.ietf.rfc8506 import *
from ...types import *


class CcRequestNumberAVP(DiameterAVP, Unsigned32Type):
    """Implementation of CC-Request-Number AVP in Section 8.2 of 
    IETF RFC 8506.

    The CC-Request-Number AVP (AVP Code 415) is of type Unsigned32.
    """
    code = CC_REQUEST_NUMBER_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             CcRequestNumberAVP.code,
                             CcRequestNumberAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class CcRequestTypeAVP(DiameterAVP, EnumeratedType):
    """Implementation of CC-Request-Type AVP in Section 8.3 of 
    IETF RFC 8506.

    The CC-Request-Type AVP (AVP Code 416) is of type Enumerated.
    """
    code = CC_REQUEST_TYPE_AVP_CODE
    vendor_id = None

    values = [
                CC_REQUEST_TYPE_INITIAL_REQUEST,
                CC_REQUEST_TYPE_UPDATE_REQUEST,
                CC_REQUEST_TYPE_TERMINATION_REQUEST,
                CC_REQUEST_TYPE_EVENT_REQUEST
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             CcRequestTypeAVP.code,
                             CcRequestTypeAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        EnumeratedType.__init__(self, data=data)


class RatingGroupAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Rating-Group AVP in Section 8.29 of 
    IETF RFC 8506.

    The Rating-Group AVP (AVP Code 432) is of type Unsigned32.
    """
    code = RATING_GROUP_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             RatingGroupAVP.code,
                             RatingGroupAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class ServiceIdentifierAVP(DiameterAVP, Unsigned32Type):
    """Implementation of Service-Identifier AVP in Section 8.28 of 
    IETF RFC 8506.

    The Service-Identifier AVP (AVP Code 439) is of type Unsigned32.
    """
    code = SERVICE_IDENTIFIER_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             ServiceIdentifierAVP.code,
                             ServiceIdentifierAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        Unsigned32Type.__init__(self, data=data)


class UserEquipmentInfoAVP(DiameterAVP, GroupedType):
    """Implementation of User-Equipment-Info AVP in Section 8.49 of 
    IETF RFC 8506.

    The User-Equipment-Info AVP (AVP Code 458) is of type Grouped.
    """
    code = USER_EQUIPMENT_INFO_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             UserEquipmentInfoAVP.code,
                             UserEquipmentInfoAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        GroupedType.__init__(self, data=data)


class UserEquipmentInfoTypeAVP(DiameterAVP, EnumeratedType):
    """Implementation of User-Equipment-Info-Type AVP in Section 8.50 of 
    IETF RFC 8506.

    The User-Equipment-Info-Type AVP (AVP Code 459) is of type Enumerated.
    """
    code = USER_EQUIPMENT_INFO_TYPE_AVP_CODE
    vendor_id = None

    values = [
                USER_EQUIPMENT_INFO_TYPE_IMEISV,
                USER_EQUIPMENT_INFO_TYPE_MAC,
                USER_EQUIPMENT_INFO_TYPE_EUI64,
                USER_EQUIPMENT_INFO_TYPE_MODIFIED_EUI64,
    ]

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             UserEquipmentInfoTypeAVP.code,
                             UserEquipmentInfoTypeAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        EnumeratedType.__init__(self, data=data)


class UserEquipmentInfoValueAVP(DiameterAVP, OctetStringType):
    """Implementation of User-Equipment-Info-Value AVP in Section 8.51 of 
    IETF RFC 8506.

    The User-Equipment-Info-Value AVP (AVP Code 460) is of type OctetString.
    """
    code = USER_EQUIPMENT_INFO_VALUE_AVP_CODE
    vendor_id = None

    def __init__(self, data):
        DiameterAVP.__init__(self, 
                             UserEquipmentInfoValueAVP.code,
                             UserEquipmentInfoValueAVP.vendor_id)
        DiameterAVP.set_mandatory_bit(self, True)
        OctetStringType.__init__(self, data=data)